/*
 *
 *
 */
#include "sim.h"

/*
 * Sets up an interrupt line and resets its counters and state.
 */
int irq_initsingle(struct irq_s *irq, const char *name, unsigned int vector)
{
  if (!irq)
    return -1;
  
  if (name)
    strcpy(irq->name, name);
  
  irq->vector = vector; 
  irq->count = 0;
  irq->precount = 0;
  irq->state = 0;

  return 0;
}

/*
 * Flag an interrupt line as high.  The vector will NOT get executed on
 * the next pass; it will only get executed when irq->precount reaches
 * the IRQWAITLEN threshold.
 *
 */
int irq_flag(int irq)
{
#ifdef PROF_IRQ
  prof_start();
#endif
#ifdef PROF_IRQFIRE
  prof_start();
#endif  
  if ((irq<0) && (irq>=IRQCOUNT))
    return -1;
  curdsp.irqs[irq].state |= IRQ_HIGH;
  curdsp.irqs[irq].count++;
  fprintf(stderr, "irq_flag: interrupt on irq %02d\n", irq);
  return 0; 
}

/*
 * Unflag interrupt and reset precount. 
 */
int irq_unflag(int irq)
{
  if ((irq<0) && (irq>=IRQCOUNT))
    return -1;
  if (curdsp.irqs[irq].state & IRQ_HIGH)
    {
      curdsp.irqs[irq].state ^= IRQ_HIGH;
      curdsp.irqs[irq].precount = 0;
    }
  return 0;
}

/*
 * This returns the highest priority interrupt that is pending,
 * assuming that the IRQs were initialized in the lowest-index-
 * is-highest-priority order.
 *
 */
int irq_checkpending(void)
{
  int i;
  for(i=0;i<IRQCOUNT;i++)
    {
      if (curdsp.irqs[i].state & IRQ_HIGH)
	return i;
    }
  return -1;
}

/*
 * Same as irq_checkpending() but only returns the irq IF its precount
 * has reached a value high enough to be allowed to execute (IRQWAITLEN).
 *
 */
int irq_checkdone(void)
{
  int i;
  for(i=0;i<IRQCOUNT;i++)
    {
      if (curdsp.irqs[i].state & IRQ_HIGH)
	{
	  if (curdsp.irqs[i].precount > IRQWAITLEN)
	    {
	      return i;
	    }
	}
    }
  return -1;
}

/*
 *
 * We use pre-counting as our way of simulating the three-cycle
 * delay between initial rise of the interrupt and the DSP's 
 * actual call of the vector.  This incremements the precounts
 * on all pending interrupts.
 *
 */
int irq_incprecount(void)
{
  int i;
  for (i=0;i<IRQCOUNT;i++)
    {
      if (curdsp.irqs[i].state & IRQ_HIGH)
	curdsp.irqs[i].precount++;
    }
  return 0;
}

/*
 * Set/get IMASK and ICNTL. Wrapped for trapping purposes.
 *
 */
void irq_setimask(unsigned int new) { curdsp.sequencer.imask = new;}
unsigned int irq_getimask(void) { return curdsp.sequencer.imask; }
void irq_seticntl(unsigned int new) {curdsp.sequencer.icntl = new; }
unsigned int irq_geticntl(void) { return curdsp.sequencer.icntl; }
void irq_setifc(unsigned int new) 
{ 
  curdsp.sequencer.ifc = new;
  if (new & BIT0) irq_unflag(11);
  if (new & BIT1) irq_unflag(10);
  if (new & BIT2) irq_unflag(9);
  if (new & BIT3) irq_unflag(8);
  if (new & BIT4) irq_unflag(7);
  if (new & BIT5) irq_unflag(6);
  if (new & BIT6) irq_unflag(5);
  if (new & BIT7) irq_unflag(4);
  if (new & BIT8) irq_flag(11);
  if (new & BIT9) irq_flag(10);
  if (new & BIT10)irq_flag(9);
  if (new & BIT11)irq_flag(8);
  if (new & BIT12)irq_flag(7);
  if (new & BIT13)irq_flag(6);
  if (new & BIT14)irq_flag(5);
  if (new & BIT15)irq_flag(4);
  return;
}
unsigned int irq_getifc(void) { return curdsp.sequencer.ifc; }

/*
 * Returns true if the irq is not disabled in imask.
 * 
 * FIXME: this shouldn't depend on the internal irq order like this.
 *
 */
int irq_checkimask(int irq)
{
  switch(irq)
    {
    case 11: return irq_getimask()&0x01; /* timer */
    case 10: return irq_getimask()&0x02; /* SPORT1 Rx */
    case  9: return irq_getimask()&0x04; /* SPORT1 Tx */
    case  8: return irq_getimask()&0x08; /* BDMA */
    case  7: return irq_getimask()&0x10; /* /IRQE */
    case  6: return irq_getimask()&0x20; /* SPORT0 Rx */
    case  5: return irq_getimask()&0x40; /* SPORT0 Tx */
    case  4: return irq_getimask()&0x80; /* /IRQL0 */
    case  3: return irq_getimask()&0x100;/* /IRQL1 */
    case  2: return irq_getimask()&0x200;/* /IRQL2 */
    case  1: /* these should never really happen */
    case  0:
      return 0; 
    }
  return 0;
}

void irq_dump(void)
{
  int i;

  fprintf(stderr, " -- IRQ Stats (imask=0x%04x, icntl=0x%04x) --\n",
	  curdsp.sequencer.imask,
	  curdsp.sequencer.icntl);
  fprintf(stderr, "%6s\t%25s\t%4s\t%4s\t%4s\t%5s\n", "IRQ", "Name", "Vctr", "Cnt", "PCnt", "State");
  for (i=0; i < IRQCOUNT; i++)
    {
      fprintf(stderr, "irq%02d:\t%25s\t%04x\t%4d\t%4d\t%5d\n", 
	      i,
	      curdsp.irqs[i].name,
	      curdsp.irqs[i].vector,
	      curdsp.irqs[i].count,
	      curdsp.irqs[i].precount,
	      curdsp.irqs[i].state);
    }
  fprintf(stderr, " --------------------------------------------\n");
  return;
}
