/*
 * From here, we handle sequencer operations.  Namely, incrementing
 * the PC, but also handling some interrupts and looping.
 *
 */

#include "sim.h"

/*
 * seq_fetch()
 *
 * Performs basically the same function as the DSP's prefetcher and
 * "next address select logic". Fills cudsp->sequencer->curinstr with
 * the next instruction to be executed and handles the associated 
 * stack fringe.
 *
 */
unsigned int seq_fetch(void)
{
  unsigned char instrbuf[MAXDECODE+1];
  unsigned int ret = 0;

  if (irq_checkdone() != -1)
    {
      /* we have IRQs we need to process */
      if ((ret=seq_callirq())<0xffff)
	return (curdsp.sequencer.pc.n = ret); /* set the new pc */
     
    }
  
  if (curdsp.sequencer.loop_stack_len)
    {
      unsigned int last,cond;
      last = loop_peekaddr();
      cond = loop_peekcond();
      
      if (curdsp.sequencer.pc.n == last+1)
	{
	  if (!testtermcond(cond))
	    {
	      /* iterate again */
	      curdsp.sequencer.pc.n = pc_peek(); 
	      curdsp.sequencer.cntr -= 1;
	    }
	  else
	    {
	      /* done, starting popping */
	      pc_pop();
	      count_pop();
	      loop_pop();
	      return curdsp.sequencer.pc.n;
	    }
	}
    }
  /* grab the instruction at (PC) */
  curdsp.sequencer.curinstr = pmgetiword(curdsp.sequencer.pc.n);
  /* show the instruction */
  prettydecode(curdsp.sequencer.curinstr, instrbuf);
  fprintf(stderr, "seq_fetch: pc = %04x\tinstr: %06x\t%s\n", 
	  curdsp.sequencer.pc.n,
	  curdsp.sequencer.curinstr, 
	  instrbuf); 
  /* execute the instruction and return new PC */
#ifdef PROF_DECODE
  prof_start();
  ret = decode();
  prof_end();
  fprintf(stderr, "seq_fetch: decode time: %8ld usec\n", prof_gettime());
#else
  ret = decode();
#endif
  if (ret!=RET_NOCHANGE)
    return (curdsp.sequencer.pc.n = ret);
  else
    return (++curdsp.sequencer.pc.n);
}

/*
 *
 * seq_callirq()
 *
 * Call interrupt _IF_ its not masked off in IMASK.
 *
 * Pushes PC (containing address of next instruction) onto the PC
 * stack and the state registers ASTAT, MSTAT, and IMASK are pushed
 * onto the state stack (in that order), and sets the main PC to the 
 * IRQ ISR vector address. When IMASK is pushed, it is reloaded with
 * a value determining whether or not nestable interrupts are allowed
 * (based on bit in ICNTL).
 *
 * We should really allow for the higher-priority-called-
 * while-in-a-lower-priorty-ISR case that the DSP does support.
 *
 * Also, some attempt should be made at emulating the fact that
 * two more standard-flow instructions get executed prior to a 
 * nop followed by the first ISR instruction.
 *
 */
unsigned int seq_callirq(void)
{
  int irq = -1;
#ifdef PROF_IRQFIRE
  prof_end();
  fprintf(stderr, "seq_fetch: IRQ fire-to-ISR latency: %8ld usec\n", prof_gettime());
#endif
  if ((irq = irq_checkdone()) < 0)
    return 0xffff;

  /* unmark the line */
  irq_unflag(irq);

  if (!irq_checkimask(irq))
    return 0xffff;

  curdsp.sequencer.pc.n--; /* back up this since the IRQ accured right beforeit*/
  /* save states... */
  pc_push();
  state_push();
  fprintf(stderr, "irq: jumping to vector at %04x\n",  curdsp.irqs[irq].vector);
  return curdsp.irqs[irq].vector;
}

/*
 * seq_rti() - return from interrupt.  Pops the top of the PC
 * stack into the PC and the top of the status stack into 
 * the status registers and returns execution to main program
 * flow.
 *
 */
int seq_rti(void)
{
  state_pop();
  pc_pop();
#ifdef PROF_IRQ
  prof_end();
  fprintf(stderr, "seq_fetch: full IRQ latency: %8ld usec\n", prof_gettime());
#endif
  return 0;
}

/*
 * Call this when you want to waste/finish a cycle.
 *
 */
int idler(void)
{     
  irq_incprecount();
  return 0;
}
